#!/usr/bin/env python
# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.
import argparse
import asyncore
import atexit
import avro.datafile
import avro.io
import avro.ipc
import avro.protocol
import cgroups
import ConfigParser
import downloader
import errno
import fcntl
import grp
import itertools
import logging
import logging.handlers
import log_tailer
import os
import parcel
import parcel_cache
import pickle
import platform
import psutil
import pyinotify
import pwd
import random
import re
import resource
import sha
import shutil
import signal
import simplejson as json
import socket
import subprocess
import sys
import tempfile
import threading
import time
import errno
import glob

try:
  from cStringIO import StringIO
except ImportError:
  from StringIO import StringIO

import supervisor.options
from supervisor.xmlrpc import Faults
from xmlrpclib import Fault

import cmf.kt_renewer
import cmf.status_server
import cmf.https
import cmf.monitor
import cmf.version
from cmf.monitor.host.filesystem_map import FilesystemMap
from cmf.supervisor_listener import SupervisordEventMonitor
from cmf.template_util import render_text
from cmf.util import unzip, tail, user_group_to_ids, register_sigquit_stack_dump, deploy_special_file, bytes_to_hex, source
from cmf.log_buffer import FixedBufferHandler
from cmf.version_detector import VersionDetector
"""
The CMF Agent starts a "status" web server, starts "supervisord",
and heartbeats regularly to the "master" server.
"""

LOG = logging.getLogger(__name__)

AGENT_PROTOCOL_VERSION = int(
  file(os.path.join(os.path.dirname(__file__), "agent_protocol_version")).read().strip())

DEFAULT_HEARTBEAT_INTERVAL_SECONDS = 5.0

# This will be used by xmlrpc.
DEFAULT_SOCKET_TIMEOUT = 30.0

DEFAULT_LOG_FILE_TAIL = 2048

# Number of heartbeats at which to compute heavier-duty
# heartbeats (like mountpoints).
RARE_HEARTBEAT_MULTIPLIER = 60

MASTER_AGENT_PROTOCOL = avro.protocol.parse(
  file(os.path.join(os.path.dirname(__file__), "master_agent.avpr")).read())
HEARTBEAT_REQUEST_SCHEMA = MASTER_AGENT_PROTOCOL.types_dict.get("HeartbeatRequest")

# CM status codes are expected to be reported in the format
# "CM_STATUS_CODE=<status code>"
CM_STATUS_CODE_MARKER = "CM_STATUS_CODE="
CM_STATUS_CODE_REGEX = CM_STATUS_CODE_MARKER + "(.*)$"
MAX_CM_STATUS_CODE_STR_LEN = len(CM_STATUS_CODE_MARKER) + 80
STATUS_CODES_PROTOCOL = avro.protocol.parse(
  file(os.path.join(os.path.dirname(__file__), "status_codes.avpr")).read())
CM_STATUS_CODE_SCHEMA = STATUS_CODES_PROTOCOL.types_dict.get("CMStatusCode")
CM_STATUS_CODES_ENV_VAR = "CM_STATUS_CODES"

MAX_RECENT_SUPERVISOR_FAILURES = 5

MAX_LOG_SIZE_BYTES = 10*1024*1024 # 10MB
LOG_BACKUP_COUNT = 5

# Number of seconds within which to log connection refused error at most once
CONN_REFUSED_THROTTLE_DURATION = 60

# Number of seconds to wait between supervisor connect attempts
SUPERVISOR_CONNECT_WAIT_TIME_SECS = 1

# Number of seconds to wait for a process to die
PROCESS_DEATH_CHECK_INTERVAL_SECS = 1

# Number of attempts to check if a process we're trying to kill has died
MAX_PROCESS_DEATH_CHECKS = 5

# Number of seconds to wait between version detection
VERSION_DETECTION_INTERVAL = 300

# Marker in a raw process object indicating that process
# was reconstructed from agent data
RECONSTRUCTED = "__reconstructed__"

# Process name for the supervisord listener.
SUPERVISOR_LISTENER_PROC = "cmflistener"

# Running processes that exit with a status not listed here will be
# automatically restarted by supervisor (if configured to do so).
SUPERVISOR_EXPECTED_EXITCODES = [0, 2]

class HeartbeatRequestMonitor(asyncore.file_dispatcher):
  """
  Dispatcher used to monitor the pipe notified by the status server when an
  explicit heartbeat request arrives.
  """

  def __init__(self, fd):
    fcntl.fcntl(fd, fcntl.F_SETFL, os.O_NONBLOCK)
    self.pipe = fd
    self.heartbeat = False
    asyncore.file_dispatcher.__init__(self, fd)

  def writable(self):
    return False

  def handle_read(self):
    while True:
      try:
        os.read(self.pipe, 32)
      except OSError, ose:
        if ose.errno == errno.EINTR:
          continue
        if ose.errno == errno.EAGAIN:
          break
        raise
    self.heartbeat = True

  def is_hearbeat_requested(self):
    """
    Returns whether a heartbeat has been requested. Resets the internal state
    so that a subsequent call returns False (unless a new request arrives).
    """
    ret = self.heartbeat
    self.heartbeat = False
    return ret


class Agent(object):
  """
  The Agent checks in with the master to keep
  the host machine in line with the processes the server
  asks it to be running.

  start() is the main loop.
  """
  def __init__(self):
    # Initialize State

    # Identifier is used to confirm that the agent is talkign to the supervisor
    # that it itself launched.  We've seen instances where the supervisor
    # doesn't die after its parent agent dies, and this causes considerable
    # consternation.
    self.identifier = "agent-%d-%d" % (os.getpid(), long(time.time()))

    # Processes, keyed by id.
    #
    # A process can either be in one of these three lists.
    # The failed processes could contain previously running
    # processes that failed to activate, or processes that
    # never successfully activated.
    self.active_processes = dict()
    self.retired_processes = dict()
    self.failed_processes = dict()

    self.shutdown_registry = []
    # Because the web server is multi-threaded, we need
    # to have some thread-safety.
    self.lock = threading.Lock()

    # Initialized by argument parsing
    self.args = None
    self.dir = None

    # Initial Heartbeat Interval
    self.heartbeat_interval = DEFAULT_HEARTBEAT_INTERVAL_SECONDS

    # For standalone mode
    self.heartbeat_output = None

    self.last_successful_heartbeat = None
    self.last_heartbeat_request = None
    self.last_heartbeat_response = None
    self.last_failed_heartbeat_request = None
    self.last_activated_symlinks = None

    self.honor_users_and_groups = None

    self.started = False

    self.last_logged_connection_refused = 0

    # For counting the number of consecutive supervisor failures
    self.recent_supervisor_failures = 0
    # For collecting metrics on the latency and failure count of
    # supervisord contact
    self.supervisor_statistics = {"failures":0, "successes":0, "latency_sec":0.0}

    # For rarely changing information that we don't want to compute every
    # heartbeat
    self.last_rare_heartbeat = 0
    self.cached_filesystem_info = None

    # For version detection
    self.last_version_detection = 0
    self.last_cm_version_detection = 0
    # Initialize to [] so that it's valid as part of a heartbeat
    self.cached_component_info = []
    self.cached_cm_component_info = None

    self.monitor_daemon = None

    # Token that can be used to verify http requests to the agent status server
    # first use of this will be for the download log file functionality
    self.token = generate_random_bytes(16)
    LOG.debug("Agent token is: %s", bytes_to_hex(self.token))

    # The agent process itself, so that we can monitor
    # our own usage.
    self.process = psutil.Process(int(os.readlink("/proc/self")))

    # Conditional so we don't blow up on OSX or whatever.
    self.dist = None
    if "dist" in dir(platform):
      dist = dict()
      dist["name"], dist["version"], dist["id"] = platform.dist()
      self.dist = dist

  def start(self):
    self.started = True

    pid = os.getpid()
    if pid == os.getpgid(pid):
      # We're the process group leader already, which is ok.
      # Typically happens when running from the command line directly.
      LOG.info("Not starting a new session.")
    else:
      os.setsid()

    atexit.register(self.stop)
    self.should_exit = False
    self.signal_caught = None
    signal.signal(signal.SIGINT, self.ready_exit_on_signal)
    signal.signal(signal.SIGTERM, self.ready_exit_on_signal)
    self.parse_arguments()
    if self.ip_address == "127.0.0.1":
      LOG.warning("Agent is running on 127.0.0.1 (localhost). This is a misconfiguration for multi-machine clusters. Check your hostname settings.")
    if "_" in self.hostname:
      raise Exception("Hostname is invalid; it contains an underscore character.")

    self.parse_config_file()
    if self.logfile != '-':
      setup_rolling_logging(self.logfile)

    # Put agent env vars starting with CMF_AGENT_ in supervisee env
    # Add/override with config.ini
    # Finally add/override with variables in command line
    if self.cmd_line_env_vars:
      for k,v in self.cmd_line_env_vars.iteritems():
        LOG.info("Adding from command line " + k + " with value: " + v)
        self.supervisee_environment[k] = v
    else:
      LOG.info("No command line vars")

    if self.master_host == None and not self.args.standalone:
      raise Exception("No server_host line was found in agent's configuration. One of --standalone or --master must be specified.")

    uid, gid = os.getuid(), os.getgid()
    user, group = pwd.getpwuid(uid)[0], grp.getgrgid(gid)[0]
    LOG.info("Agent starting as pid %s user %s(%d) group %s(%d)." % (pid, user, uid, group, gid))

    if uid != 0:
      LOG.info("Because agent not running as root, all processes will run with "
               "current user.")
      self.honor_users_and_groups = False
    else:
      self.honor_users_and_groups = True

    self.mkworkdir()

    self.mkabsdir(self.parcel_dir, user=user, group=group, mode=0755)
    self.repo = parcel.Repository(self.parcel_dir)
    self.repo.refresh()
    try:
      inotify = self.repo.configure_inotify()
      self.register_for_shutdown(inotify.stop)
    except:
      LOG.exception("Failed to configure inotify. Parcel repository will not auto-refresh.")
      inotify = None

    parcel_cache_dir = os.path.join(os.path.dirname(self.parcel_dir), "parcel-cache")
    self.mkabsdir(parcel_cache_dir, user=user, group=group, mode=0755)
    self.downloader = downloader.Downloader(parcel_cache_dir, 1)
    self.register_for_shutdown(self.downloader.cleanUp)
    self.cache = parcel_cache.ParcelCache(parcel_cache_dir, self.repo, self.downloader)
    self.repo.set_parcel_added(self.cache.prune_parcel_download_history)

    # Initialize cgroups state.
    cgroups_dir = self.mkworkdir("cgroups")
    self.cg_manager = cgroups.CGroupManager(cgroups_dir)
    if self.cg_manager.capabilities:
      LOG.info("Found cgroups capabilities: %s" %
               (self.cg_manager.capabilities,))
      self.register_for_shutdown(self._cgroups_conditional_unmount)
    else:
      LOG.info("No cgroups capabilities found")
    # Tracks whether resource management is enabled; independent of whether
    # cgroups are actually mounted.
    self.rm_enabled = False

    self.tailer = log_tailer.LogTailer(self.host_id, self.hostname)
    self.register_for_shutdown(self.tailer.shutdown)

    # Create the named pipe that will be used to read events from supervisord,
    # if supervisord_events are enabled. Eager heartbeats and detection of
    # unexpected exits rely on supervisord_events being enabled.
    # The event fifo is kept in the agent's
    # working directory, and is written to the supervisord configuration file.
    # This means that if the agent needs to be re-started, it should use the
    # same log directory, or supervisord will have to be re-started too. (Note
    # that sending SIGHUP to supervisord makes it re-start all processes being
    # monitored, which is not what we want.)
    self.event_pipe_path = os.path.join(self.dir, "events")
    if os.path.exists(self.event_pipe_path):
      os.unlink(self.event_pipe_path)
    if self.supervisord_events:
      LOG.info("Setting up supervisord event monitor.")
      os.mkfifo(self.event_pipe_path, 0600)
      self.event_mon = SupervisordEventMonitor(self.event_pipe_path)
      self.register_for_shutdown(self.event_mon.close)

    self.fs_map = FilesystemMap(self.monitored_nodev_filesystem_types,
                                poll_timeout=self.filesystem_map_poll_timeout_seconds,
                                join_timeout=self.filesystem_map_join_timeout_seconds,
                                tolerance=self.filesystem_map_cache_tolerance)

    # Create the monitoring daemon so that any supervised process can create
    # its monitor.
    LOG.info("Using metrics_url_timeout_seconds of %f" % (self.metrics_url_timeout_seconds,))
    LOG.info("Using task_metrics_timeout_seconds of %f" % (self.task_metrics_timeout_seconds,))
    LOG.info("Using max_collection_wait_seconds of %f" % (self.max_collection_wait_seconds,))
    self.monitor_daemon = cmf.monitor.MonitorDaemon(self.hostname,
                                                    self.host_id,
                                                    self.fs_map,
                                                    self)

    LOG.info("Supervised processes will inherit the following environment: %s" % (os.environ,))
    LOG.info("Supervised processes will add the following their environment environment: %s" % (self.supervisee_environment,))

    self.find_or_start_supervisor()

    # To use eager heartbeats, create a pipe that the status server will use
    # to notify the main thread when an explicit heartbeat request arrives.
    wakeup_fd = os.pipe()
    self.hb_req_monitor = HeartbeatRequestMonitor(wakeup_fd[0])
    self.register_for_shutdown(self.hb_req_monitor.close)
    wakeup_fd = wakeup_fd[1]

    self.url = cmf.status_server.start_status_server(
       self, self.hostname, self.agent_httpd_port, wakeup_fd)

    # Start the monitoring daemon, after we have the supervisor info
    self.monitor_daemon.start()

    LOG.warning("Setting default socket timeout to %d!" % (DEFAULT_SOCKET_TIMEOUT,))
    socket.setdefaulttimeout(DEFAULT_SOCKET_TIMEOUT)

    # This is the main loop
    should_heartbeat = True
    while not self.should_exit:
      if should_heartbeat:
        last = time.time()
        try:
          # The heartbeat mechanism is broken down into
          # three methods.
          heartbeat = self.prepare_heartbeat()
          timed_heartbeat_response = self.send_heartbeat(heartbeat)
          heartbeat_response = timed_heartbeat_response.get_response()
          if heartbeat_response is not None:
            self.handle_heartbeat_response(timed_heartbeat_response)

            self.last_heartbeat_request = heartbeat
            self.last_heartbeat_response = heartbeat_response
            self.last_successful_heartbeat = time.time()
        except:
          LOG.exception("Caught unexpected exception in main loop.")

      try:
        self.tailer.process_events()
      except:
         LOG.exception("Error processing tailer events.")

      # Check for inotify events
      if inotify != None and inotify.check_events():
        inotify.read_events()
        inotify.process_events()

      diff = time.time() - last
      if self.supervisord_events:
        asyncore.loop(max(0, self.heartbeat_interval - diff),
                      use_poll=True,
                      count=1)
      elif diff < self.heartbeat_interval:
        time.sleep(max(0, self.heartbeat_interval - diff))

      if self.supervisord_events:
        events = self.event_mon.consume_events()
      else:
        events = []

      hb_requested = self.hb_req_monitor.is_hearbeat_requested()
      if events or hb_requested:
        should_heartbeat = True
        LOG.debug("Sending eager heartbeat (%d events, explicit request = %s).",
                  len(events), hb_requested and "true" or "false")
      else:
        diff = time.time() - last
        should_heartbeat = self.heartbeat_interval <= diff

      if self.supervisord_events:
        self.handle_unexpected_process_exits(events)

    self.stop()
    LOG.info("Agent exiting; caught signal %s" % (self.signal_caught,) )
    sys.exit(0)

  def _cgroups_conditional_unmount(self):
    """Unmount mounted cgroups roots only if no active processes are using
       cgroups."""
    if self._has_at_least_one_cgroup():
      LOG.info("At least one outstanding cgrouped process; retaining cgroup mounts")
      return
    else:
      LOG.info("No cgrouped processes running; unmounting any cgroup roots")
      self.cg_manager.disable()

  def _host_heartbeat(self):
    """Populates dict data with host heartbeat info."""
    status = dict()
    status["host_name"] = self.hostname
    status["ip_address"] = self.ip_address
    status["agent_url"] = self.url
    status["agent_token"] = self.token

    status["num_cores"] = psutil.NUM_CPUS

    now = time.time()
    if self.cached_filesystem_info is None or \
       (now - self.last_rare_heartbeat >=
        RARE_HEARTBEAT_MULTIPLIER * self.heartbeat_interval):
        self.fs_map.refresh()
        self.cached_filesystem_info = self.fs_map.get_info_for_scm_heartbeat()
        self.last_rare_heartbeat = now

    status["filesystem_info"] = self.cached_filesystem_info["filesystem_info"]
    status["component_info"] = self.cached_component_info
    status["parcel_info"] = self.repo.products
    status["parcel_err"] = self.repo.errors
    status["distro_info"] = self.dist
    status["cgroup_capabilities"] = self.cg_manager.capabilities

    # Because the agent is multi-threaded, we avoid
    # reading the logs, since that may require synchronization.
    status["agent_log_tail"] = ""

    stats = dict()
    stats["load_avg"] = list(os.getloadavg())
    stats["physical_memory_total"] = psutil.TOTAL_PHYMEM
    cached = psutil.cached_phymem()
    buffers = psutil.phymem_buffers()
    really_used = psutil.used_phymem() - cached - buffers
    stats["physical_memory_used"] = really_used
    stats["virtual_memory_total"] = psutil.total_virtmem()
    stats["virtual_memory_used"] = psutil.used_virtmem()
    stats["mountpoint_available_space_bytes"] = \
        self.cached_filesystem_info["mountpoint_available_space_bytes"]
    stats["parcel_downloads"] = self.downloader.get_current_downloads()

    return status, stats

  def monitoring_supervisord_events(self):
    return self.supervisord_events

  def get_supervisor_data(self, count_failures=False):
    data = None
    try:
      self.lock.acquire()
      try:
        sup = self.supervisor_client.supervisor

        supervisor_latency_start = time.time()
        supervisor_info = sup.getAllProcessInfo()
        supervisor_latency_end = time.time()
        data = dict( (v["name"], v) for v in supervisor_info )

        latency = supervisor_latency_end - supervisor_latency_start
        self.supervisor_statistics["latency_sec"] += latency
        self.supervisor_statistics["successes"] += 1
        # Reset failure count
        self.recent_supervisor_failures = 0

      except:
        # The XML-RPC Proxy tends not to recover from errors, so we refresh it.
        self.configure_supervisor_clients()
        LOG.exception("Could not contact supervisor.")

        self.supervisor_statistics["failures"] += 1
        # If we keep failing long enough, it's time to give up.
        if count_failures:
          self.recent_supervisor_failures += 1
          if self.recent_supervisor_failures > MAX_RECENT_SUPERVISOR_FAILURES:
            LOG.error("Failed to contact supervisor after %u attempts. Agent " \
                        "will exit." % self.recent_supervisor_failures)
            self.stop()
            sys.exit(-1)

    finally:
      self.lock.release()

    # Anytime we get new supervisor data, we should update any process monitors
    # that we have.
    if data is not None:
      self.update_monitors_from_supervisor_data(data)

    return data

  def prepare_heartbeat(self):
    # AVRO objects are dict-based
    request = dict()
    request["version"] = AGENT_PROTOCOL_VERSION
    request["host_id"] = self.host_id

    # Get host-level data
    request["status"] = dict()
    request["status"]["host"], request["host_stats"] = self._host_heartbeat()

    # Gather supervisor data
    supervisor_data = self.get_supervisor_data(True)

    # Let's work through active processes
    process_statuses = []
    process_stats = []

    # We combine both lists since we want to report back to the sever
    # about currently active processes, but also processes that failed
    # to activate.
    ids_seen = set()
    for process_id, process in itertools.chain(self.active_processes.iteritems(),
                                               self.failed_processes.iteritems()):
      if process_id in ids_seen:
        # This should never happen.
        LOG.warning("Duplicate process id: %s" % (process_id,));
      ids_seen.add(process_id)
      # If we've never got a HB response, then any processes
      # we have are reconstructed from an old agent and can't
      # be sent.
      if process.raw.get(RECONSTRUCTED) is None:
        status, stats = process.make_status(supervisor_data)
        process_statuses.append(status)
        process_stats.append(stats)

    request["status"]["processes"] = process_statuses
    request["process_stats"] = process_stats

    md = sha.new(pickle.dumps(request["status"]))
    request["status_hash"] = md.digest()

    cpu_total = 0.0
    try:
      # Shouldn't fail, but, if it does, let's not
      # worry about it.
      cpu_times = self.process.get_cpu_times()
      cpu_total += cpu_times.user + cpu_times.system
    except:
      LOG.exception("Exception getting agent cpu times.")

    try:
      # If the supervisor process is gone, this
      # won't work, and we don't want that to prevent
      # heartbeating.
      cpu_times = self.supervisor_psutil_process.get_cpu_times()
      cpu_total += cpu_times.user + cpu_times.system
    except:
      LOG.exception("Exception getting supervisor cpu times.")

    request["total_cpu"] = cpu_total

    if self.last_heartbeat_response is not None:
      request["last_response_hash"] = self.last_heartbeat_response["data_hash"]

      if request["status_hash"] == self.last_heartbeat_response["last_request_hash"]:
        LOG.debug("Matched request hashes. Not sending status")
        del request["status"]
      else:
        LOG.debug("Unmatched request hashes. Sending status")
    else:
      request["last_response_hash"] = ""

    # Because heartbeating sends log information to the master,
    # including the heartbeat in the log message triggers
    # recursively larger and larger heartbeats.
    return request

  def send_heartbeat(self, heartbeat):
    ts_agent_send = None
    ts_agent_recv = None
    ts_master_send = None
    ts_master_recv = None
    if self.args.standalone:
      # Write heartbeats to a file
      if not self.heartbeat_output:
        self.heartbeat_output = avro.datafile.DataFileWriter(
          file(os.path.join(self.dir, "heartbeats.avro"), "w"),
          avro.io.DatumWriter(),
          HEARTBEAT_REQUEST_SCHEMA)
      self.heartbeat_output.append(heartbeat)
      self.heartbeat_output.flush()
      # Read expected processes
      datum_reader = avro.io.DatumReader()
      dfr = avro.datafile.DataFileReader(file(self.args.standalone), datum_reader)
      response = dfr.next()
    else:
      try:
        if self.requestor is None:
          if self.use_tls:
            transceiver = cmf.https.HTTPSTransceiver(self.master_host,
                                                     self.master_port,
                                                     self.verify_cert_file,
                                                     self.client_key_file,
                                                     self.client_cert_file,
                                                     self.client_keypw_file)
          else:
            transceiver = avro.ipc.HTTPTransceiver(self.master_host,
                                                   self.master_port)
          self.requestor = avro.ipc.Requestor(MASTER_AGENT_PROTOCOL,
                                              transceiver)
        LOG.debug("Sending heartbeat to master.")
        assert self.requestor is not None
        ts_agent_send = long(time.time() * 1000)
        response = self.requestor.request('heartbeat', dict(request=heartbeat))
        ts_agent_recv = long(time.time() * 1000)
        # Take the timestamps from the response object regardless of whether or not
        # the old response object is used.
        ts_master_send = response.get('ts_send')
        ts_master_recv = response.get('ts_recv')
        LOG.debug("Received response from master: %s" % response)
        if (self.last_heartbeat_response is not None and
            (response["data_hash"] ==
             self.last_heartbeat_response["data_hash"])):
           LOG.debug("Matched response hashes. Using previous response")
           self.last_heartbeat_response["last_request_hash"] = \
             response["last_request_hash"]
           response = self.last_heartbeat_response
        else:
           LOG.debug("Unmatched response hashes. Using new response")
      except:
        self.log_heartbeat_exception("Heartbeating to %s:%s failed." % (self.master_host, self.master_port))
        self.last_failed_heartbeat_request = heartbeat
        self.requestor = None
        response = None
    return TimedHeartbeatResponse(response, ts_agent_send, ts_agent_recv,
                                  ts_master_send, ts_master_recv)

  def log_heartbeat_exception(self, message):
    ex = sys.exc_info()[1];
    if (ex is not None and hasattr(ex, 'errno') and ex.errno == errno.ECONNREFUSED):
      elapsed_time = time.time() - self.last_logged_connection_refused
      if (elapsed_time > 0 and elapsed_time <= CONN_REFUSED_THROTTLE_DURATION):
        return
      else:
        self.last_logged_connection_refused = time.time()
    LOG.exception(message)

  def _has_at_least_one_cgroup(self):
    for active_process in self.active_processes.values():
      if active_process.has_cgroups:
        return True
    return False

  def handle_heartbeat_response(self, timed_heartbeat_response):
    assert timed_heartbeat_response is not None
    assert timed_heartbeat_response.get_response() is not None

    try:
      self.lock.acquire()
      try:
        if not self.monitor_daemon is None:
          self.monitor_daemon.record_heartbeat_timestamps(
              timed_heartbeat_response.get_ts_agent_send(),
              timed_heartbeat_response.get_ts_agent_recv(),
              timed_heartbeat_response.get_ts_master_send(),
              timed_heartbeat_response.get_ts_master_recv())

        response = timed_heartbeat_response.get_response()["data"]
        self.heartbeat_interval = response["heartbeat_interval"]

        # Update our monitor daemon with the latest firehose information. Note
        # that if no information is specified, this will turn off reporting to
        # the firehose.
        firehoses = response.get("firehoses", [])
        self.monitor_daemon.update_firehoses(firehoses)

        self.monitor_daemon.enabled_metric_reporters = \
            response.get("enabled_metric_reporters", [])

        host_config_data = response.get("host_collection_config_data")
        if not host_config_data is None:
          host_config = dict()
          for c in host_config_data:
            if c.has_key('config_name') and c.has_key('config_value'):
              host_config[c['config_name']] = c['config_value']
          self.monitor_daemon.update_host_config(host_config)
        else:
          LOG.debug("Heartbeat response did not contain host collection config data.")

        active_parcels = response.get("active_parcels")

        # (De)Activate system symlinks for parcels
        new_symlink_parcels = dict()
        if response.get("create_parcel_symlinks"):
          new_symlink_parcels = active_parcels.copy()

        old_symlink_parcels = dict()
        if self.last_heartbeat_response is not None:
          old_response = self.last_heartbeat_response["data"]
          if old_response.get("create_parcel_symlinks"):
            old_symlink_parcels = self.last_activated_symlinks
        else:
          # Remove all symlinks in case we're immediately told to discard some
          # parcels
          for parcels in self.repo.products.values():
            for parcel in parcels.values():
              self.repo.configure_symlinks(parcel, False)

        # Filter out any parcels from symlinks consideration that are not yet
        # available. We do this to avoid trying to activate any parcels too early
        # By also removing these entries from the old symlink map we avoid any
        # scenario where deactivate might accidentally get called
        LOG.debug("active_parcels: " + str(active_parcels))
        for name, version in active_parcels.items():
          if not (name in self.repo.products and version in self.repo.products[name]):
            # We just delete the name as the active_parcels map never contains
            # multiple versions of the same product
            del new_symlink_parcels[name]
            if name in old_symlink_parcels:
              del old_symlink_parcels[name]
        LOG.debug("new_symlink_parcels: " + str(new_symlink_parcels))
        LOG.debug("old_symlink_parcels: " + str(old_symlink_parcels))

        self.repo.configure_all_symlinks(old_symlink_parcels, new_symlink_parcels)
        self.last_activated_symlinks = new_symlink_parcels

        if response.get("server_manages_parcels"):
          server_managed_parcels = response.get("server_managed_parcels")
          LOG.debug("Server Managed Parcels: %s", str(server_managed_parcels))
          self.cache.update(server_managed_parcels)

        # Extract the CDH component versions based on the reported active CDH parcel
        cdh_parcel_components = None
        if 'CDH' in active_parcels:
          LOG.debug("Getting component info for active CDH parcel")
          cdh_parcel_components = self.repo.get_components_for_cdh(active_parcels['CDH'])
          if 'IMPALA' in active_parcels:
            LOG.debug("Getting component info for active Impala parcel")
            cdh_parcel_components += self.repo.get_components_for_impala(active_parcels['IMPALA'])

        # Calculate the CM component versions
        now = time.time()
        if (self.cached_cm_component_info is None or
            (now - self.last_version_detection >= VERSION_DETECTION_INTERVAL)):
          LOG.debug("Populating cm component version cache from packages")
          self.cached_cm_component_info = \
            VersionDetector.get_agent_component_versions(self.supervisee_environment)
          self.last_cm_version_detection = now

        # Calculate the CDH component versions
        if cdh_parcel_components is not None:
            LOG.debug("Populating cdh component version cache from active parcel")
            self.cached_component_info = cdh_parcel_components + self.cached_cm_component_info
        elif (self.cached_component_info == [] or
             (now - self.last_version_detection >= VERSION_DETECTION_INTERVAL)):
          LOG.debug("Populating cdh component version cache from system packages")
          self.cached_component_info = \
            VersionDetector.get_component_versions(self.supervisee_environment)
          self.cached_component_info += self.cached_cm_component_info
          self.last_version_detection = now

        logs_and_rules_config = response.get("log_tailing_config")
        eventserver_host = response.get("eventserver_host")
        eventserver_port = response.get("eventserver_port")
        if logs_and_rules_config is not None:
          try:
            self.tailer.set_config(logs_and_rules_config)
            self.tailer.set_eventserver_config(eventserver_host, eventserver_port)
          except:
            LOG.exception("Error setting tailer configuration.")
        else:
          LOG.debug("Heartbeat response did not contain log tailing configuration.")

        # Mount/unmount cgroup subsystems. Process cgroup state is taken
        # into consideration:
        # 1. If a process has cgroups, we mount so that when the process
        #    dies we can properly destroy its cgroups.
        # 2. Likewise, we only unmount when all cgroups are gone, even if
        #    the server has disabled RM in the interim.
        #
        # In the interest of performance, this won't remount subsystems that
        # were unmounted by an outside party.
        self.rm_enabled = response.get("rm_enabled")
        if self.rm_enabled or self._has_at_least_one_cgroup():
          self.cg_manager.enable()
        elif not self._has_at_least_one_cgroup():
          self.cg_manager.disable()

        # Remove past processes.  Ignore already present orders.  Add new ones.
        new_processes = dict( (pr["id"], pr) for pr in response["processes"])
        assert len(new_processes) == len(set(new_processes.keys())), "Processes must have unique ids."

        # Clear failed processes that the server no longer cares about,
        # and don't reactivate failed processes.
        for process_id in self.failed_processes.keys():
          if process_id not in new_processes:
            # Error information is no longer useful,
            # since the server no longer recognizes this id.
            del self.failed_processes[process_id]
          else:
            # This process failed in a fatal way already;
            # we don't want to activate it again
            del new_processes[process_id]

        for id, active_process in self.active_processes.items():
          if id not in new_processes:
            try:
              active_process.deactivate(True)
              del self.active_processes[id]
              self.retired_processes[id] = active_process
              LOG.info("Retiring process %s" % (active_process,))
              # Lets clear the error just in case we ever need to use retired processes
              active_process.clear_error()
            except RetryableProcessException:
              # We'll try again in the next heartbeat cycle, provided the server
              # still omits it from the heartbeat response.
              LOG.exception("Failed to retire %s - Will retry soon." % (active_process,))
            except Exception, e:
              # We don't retry these exceptions
              LOG.exception("Failed to retire %s" % (active_process,))
              # This will be in failed processes for one heartbeat,
              # but then it will disappear entirely because
              # of "del self.failed_processes[...]" above.
              self.add_failed_process(active_process, e)
          else:
            # Subtract from new_process immediately, in case an exception is
            # raised. If we don't do this now, we may end up orphaning the
            # existing process without actually deactivating it.
            new_process = new_processes[id]
            del new_processes[id]
            try:
              changed = False
              # Detect this change quickly, since we have processes with
              # "fake_raw" data upon agent restart.
              if active_process.raw != new_process:
                LOG.info("Process with same id has changed: %s." % (active_process.name))
                active_process.raw = new_process
                changed = True
              active_process.raw["run_generation"] = new_process["run_generation"]
              if new_process["running"]:
                active_process.activate()
              else:
                active_process.deactivate(False)
              if changed:
                #
                # The process specifies which files can be refreshed.
                #
                if active_process.activated:
                  LOG.info("Refreshing files for process %s: %s" % (active_process.name, active_process.raw["refresh_files"]))
                  active_process.write_process_conf(active_process.raw["refresh_files"])
                  active_process.borrow_keytab()
                  active_process.refresh_monitor()

              # Everything is good with the active process.
              active_process.clear_error()
            except RetryableProcessException:
              # We'll try again in the next heartbeat cycle, provided the
              # server still updates it in the heartbeat response.
              LOG.exception("Failed to update %s - Will retry soon." % (active_process,))
            except Exception, e:
              # We don't retry these exceptions
              LOG.exception("Failed to update %s" % (active_process,))
              self.add_failed_process(active_process, e)

        # We already subtracted the pre-existing ones from new_processes.
        # These are the truly new processes.
        for raw in new_processes.values():
          new_process = None
          try:
            new_process = Process(self, raw)
            if raw["running"]:
              new_process.activate()
            self.active_processes[raw["id"]] = new_process
          except RetryableProcessException:
            LOG.exception("Failed to activate %s - Will retry soon." % (raw,))
          except Exception, e:
            LOG.exception("Failed to activate %s" % (raw,))
            self.add_failed_process(new_process, e)
      except:
        # Worst case exception handler
        LOG.exception("Failed to handle Heartbeat Response: %s" % response)
    finally:
      self.lock.release()

  def handle_unexpected_process_exits(self, events):
    ''' Looks through the received events for unexpected process exits
        and updates the corresponding monitors unexpected exit state.'''

    processes = self.active_processes.values() + self.failed_processes.values()
    unexpected_process_exits = dict()

    for event in events:
      if event.properties["eventname"] == "PROCESS_STATE_EXITED":
        process_pid = event.payload_properties["pid"]
        if process_pid is None:
          LOG.warn("Process exited event payload '%s' does not have a "\
                     "'pid' field - ignoring event", event.payload)
          continue

        process_name = event.payload_properties["processname"]
        if process_name is None:
          LOG.warn("Process exited event payload '%s' for PID '%s' " \
                     "does not have a 'processname' field - " \
                     "ignoring event", event.payload, process_pid)
          continue

        expected = event.payload_properties["expected"]
        if expected is None:
          LOG.warn("Process exited event payload '%s' for PID '%s' " \
                     "associated with process '%s' does not have a " \
                     "'expected' field - ignoring event",
                   event.payload, process_pid, process_name)
          continue

        # An 'expected' value of 0 indicates that the process has
        # exited unexpectedly
        if expected == str(0):
          LOG.info("PID '%s' associated with process '%s' " \
                     "with payload '%s' exited unexpectedly",
                    process_pid, process_name, event.payload)
          unexpected_process_exits[process_name] = process_pid

    for process in processes:
      if unexpected_process_exits.has_key(process.name):
        if process.monitor is None:
          LOG.warn("Process exited event for process '%s' does not have a " \
                     "monitor to handle event - the event will be ignored.",
                   process.name)
          continue
        process.monitor.record_unexpected_exit()

  def add_failed_process(self, process, error):
    ''' Adds a failed process and also removes
        it from the active process list
    '''
    if process is not None:
      process.register_error(error)
      self.failed_processes[process.id] = process
      if process.id in self.active_processes:
        del self.active_processes[process.id]

  def update_monitor_process_information(self):
    self.update_monitors_from_supervisor_data(self.get_supervisor_data())

  def update_monitors_from_supervisor_data(self, supervisor_data):
    if supervisor_data is None:
      return

    # Only refresh the monitor if there's actual supervisor data to act on.
    try:
      self.lock.acquire()
      for process in self.active_processes.itervalues():
        if process.monitor is None:
          continue
        data = supervisor_data.get(process.name)
        pid = None
        if data is not None and data["statename"] == "RUNNING":
          pid = data["pid"]
        process.monitor.refresh_pid(pid)
    finally:
      self.lock.release()

  def mkworkdir(self, *args, **kwargs):
    """
    Creates a directory that's a subdirectory of the agent
    directory.

    Args are paths to be concatenated with the
    agent working directory.

    The only valid keyword args are user and group.
    """
    assert set(["user", "group"]).issuperset(kwargs.keys())
    user = kwargs.get("user")
    group = kwargs.get("group")

    d = os.path.join(self.dir, *args)
    created = self.mkabsdir(d, user=user, group=group, mode=0751)
    if not created:
      LOG.info("Re-using pre-existing directory: %s" % (d,))
    return d

  def mkabsdir(self, path, user, group, mode=None):
    """
    Creates a directory; path is absolute.

    Returns true if a directory was created.
    """
    if os.path.isdir(path):
      return False

    # Note that makedirs is hierarchical, but we only change permissions
    # on the leaf directory.
    os.makedirs(path)
    LOG.info("Created %s" % (path,))
    if self.honor_users_and_groups and user is not None and group is not None:
      uid, gid = user_group_to_ids(user, group)
      LOG.info("Chowning %s to %s (%s) %s (%s)" % (path, user, uid, group, gid))
      os.chown(path, uid, gid)
    if mode is not None:
      LOG.info("Chmod'ing %s to 0%o" % (path, mode))
      try:
        os.chmod(path, mode)
      except:
        LOG.warning("Chmod failed (%s)" % (path, ))
    return True

  def find_or_start_supervisor(self):
    # Because I am wary of multi-threading issues,
    # we spin off a subprocess with supervisor.

    # Explicitly create the working directories to ensure they
    # exist and have sufficient permissions to be usable.
    self.mkworkdir("process")
    self.supervisor_dir = self.mkworkdir("supervisor")
    self.supervisor_include_path = self.mkworkdir("supervisor", "include")
    # TODO: should be configurable
    self.supervisor_url = "http://%s:%s/" % (self.hostname, self.supervisord_httpd_port)

    # Check for an already running supervisor. We may have deliberately left
    # it running to avoid shutting down processes, or the agent may have
    # crashed. In either case, we want to reconnect and reuse it.
    try:
      # We need the running supervisor's id, and this call has the nice
      # side-effect that if it fails, we don't have a reusable supervisor.
      self.configure_supervisor_clients()
      self.get_supervisor_process_info()
      LOG.info("Connecting to previous supervisor: %s." % (self.identifier, ))

      # Find the existing processes and re-construct the bare minimum Agent
      # state for them. This allows us to be aware of them and avoid treating
      # them as brand-new processes when the first heatbeat response arrives.
      #
      # Note that these 'fake' process objects are not complete enough to
      # use for a heartbeat.
      existing_processes = None
      while existing_processes == None:
        existing_processes = self.get_supervisor_data(True)

      for full_id, existing_process in existing_processes.items():
        # Our event listener doesn't conform to the CM process naming
        # convention. It's also not a process monitored by CM itself,
        # so it's best just to ignore it.
        if full_id == SUPERVISOR_LISTENER_PROC:
          continue

        split_id = full_id.split('-', 1)
        process_id = int(split_id[0])
        fake_raw = dict()
        fake_raw["id"] = process_id
        fake_raw["name"] = split_id[1]
        fake_raw[RECONSTRUCTED] = True
        LOG.info("Pre-existing process: %s" % full_id)
        new_process = Process(self, fake_raw)
        # We need to deduce the correct activated state from the
        # process state.
        # RUNNING: Obviously, this should be activated
        # STARTING: To get into the STARTING state, it must have been activated
        # BACKOFF: To get into this state, it must have been STARTING
        # FATAL: Less obvious, but this state only occurs when BACKOFF fails
        # EXITED: Least obvious. This occurs when the process terminates itself.
        #         When we see such a process, we don't know if the server is
        #         aware of this state. If we mark it as activated, and the server
        #         thinks its not running, it will ask us to deactivate it, which
        #         is a no-op. If the server thinks it is running, it will not
        #         ask us to change anything. Then it will see the EXITED state
        #         and mark the process as deactivated, and we're back in sync.
        # STOPPING and STOPPED: These map to deactivated.
        # UNKNOWN: Could be anything but let's assume deactivated.
        if existing_process['statename'] == 'RUNNING' or \
           existing_process['statename'] == 'STARTING' or \
           existing_process['statename'] == 'BACKOFF' or \
           existing_process['statename'] == 'FATAL' or \
           existing_process['statename'] == 'EXITED':
          new_process.activated = True
        new_process.dir = os.path.join(self.dir, "process", new_process.name)
        new_process.logdir = os.path.join(new_process.dir, "logs")
        # If we reconnected to an older supervisor, it might not have this.
        new_process.has_cgroups = bool(existing_process.get('has_cgroups'))
        self.active_processes[process_id] = new_process

      self.register_supervisor_for_shutdown()
      # We've found the previous supervisor, so we're done here.
      return
    except:
      LOG.info("Failed to connect to previous supervisor.")

    #
    # We don't have an existing supervisor, so let's start one.
    #

    # Clear the include path (in case there are remnants):
    for f in os.listdir(self.supervisor_include_path):
      p = os.path.join(self.supervisor_include_path, f)
      LOG.info("Deleting stale supervisor include %s." % p)
      os.remove(p)

    umask = os.umask(0177)
    config = file(os.path.join(self.supervisor_dir, "supervisord.conf"), "w")

    # First, we create the config
    config.write(render_text("supervisord.conf", data={
      'supervisor_include_path': self.supervisor_include_path,
      'supervisord_httpd_port': self.supervisord_httpd_port,
      'supervisord_username': str(random.randint(sys.maxint / 2, sys.maxint)),
      'supervisord_password': str(random.randint(sys.maxint / 2, sys.maxint)),
      'supervisord_logfile': os.path.join(self.logdir, "supervisord.log"),
      'identifier': self.identifier,
      'supervisord_events': self.supervisord_events,
      'listener_name': SUPERVISOR_LISTENER_PROC,
      'cmf_python' : sys.executable,
      'cmf_agent_path' : os.path.dirname(__file__),
      'cmf_agent_fifo' : self.event_pipe_path,
      'cmf_listener_log' : os.path.join(self.logdir, "cmf_listener.log"),
    }))
    config.close()
    os.umask(umask)

    # Now start the guy
    p = subprocess.Popen(
      args=[self.supervisord_path],
      stdin=None,
      stdout=None,
      stderr=None,
      close_fds=True,
      cwd=self.supervisor_dir)
    # Supervisor forks itself off immediately, so we wait on the
    # subprocess to avoid keeping a defunct process around.
    p.wait()

    self.connect_to_new_supervisor()
    self.register_supervisor_for_shutdown()

  def connect_to_new_supervisor(self):
    """
    Connect to the newly launched supervisor and try to get its process info
    If we fail to connect to the supervisor the agent exits.
    """
    connected = False
    for attempt in range(0, MAX_RECENT_SUPERVISOR_FAILURES):
      try:
        # give the supervisor a chance to start
        time.sleep(SUPERVISOR_CONNECT_WAIT_TIME_SECS)
        LOG.info("Trying to connect to newly launched supervisor (Attempt %d)" \
          % (attempt + 1,))
        self.configure_supervisor_clients()
        self.get_supervisor_process_info()
      except:
        LOG.exception("Failed! trying again in %s second(s)" %
          (SUPERVISOR_CONNECT_WAIT_TIME_SECS))
        continue

      connected = True
      LOG.info("Successfully connected to supervisor")
      break

    if not connected:
      LOG.error("Failed to connect to newly launched supervisor. " \
        "Agent will exit")
      self.stop()
      sys.exit(-1)


  def get_supervisor_process_info(self):
    """
    Retrieves the supervisor ID and PID. Throws an exception if it fails to
    connect to the supervisor.
    """
    self.identifier = self.supervisor_client.supervisor.getIdentification()
    self.supervisor_pid = self.supervisor_client.supervisor.getPID()
    self.supervisor_psutil_process = psutil.Process(self.supervisor_pid)

  def register_supervisor_for_shutdown(self):
    """
    Register the supervisor for shutdown using kill_supervisor which only kills
    it if it is not managing any processes.
    """
    # Helper for killing. Popen.kill(), alas, doesn't exist until 2.6
    def kill_supervisor(pid):
      num_running = 0
      for id in self.active_processes:
        if self.active_processes[id].activated:
          num_running += 1
      if num_running == 0:
        LOG.info("No processes are being managed; Supervisor will shutdown.")
      else:
        LOG.info("%u processes are being managed; Supervisor will continue to run." %
                 num_running)
        return

      # Now shut it down
      LOG.info("Shutting down supervisord, pid %s" % (pid,))
      try:
        success = self.supervisor_client.supervisor.shutdown()
        if success:
          LOG.info("Successfully shutdown supervisord, pid %s" % (pid,))
          return
        else:
          LOG.error("Failed to shutdown supervisord via rpc, trying to kill it...")
      except:
        LOG.exception("Failed to shutdown supervisord, pid %s. Killing it!" % (pid,))

      # Failed to shutdown, let's try to kill it
      self.kill_process(pid)

    self.register_for_shutdown(lambda: kill_supervisor(self.supervisor_pid))

  def configure_supervisor_clients(self):
    """
    Configure the supervisor client.
    We could do the XMLRPC configuration manually, but
    why bother if there's helper code to do it for us!
    """
    supervisor_options = supervisor.options.ClientOptions()
    supervisor_options.realize(args=["-c", os.path.join(self.supervisor_dir, "supervisord.conf")])
    self.supervisor_client = supervisor_options.getServerProxy()
    self.supervisor_ctl = supervisor.supervisorctl.Controller(supervisor_options)

  def _default_supervisord_path(self):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..", "build", "env", "bin", "supervisord")

  def add_env_vars_with_prefix_to_supervisee_environment(self, env_prefix):
    """ Add environment variables that start with env_prefix to
        supervisee's environment.
    """
    LOG.info("Adding env vars that start with %s" %(env_prefix))
    for key, value in os.environ.iteritems():
      if key.startswith(env_prefix):
        env_key = key[len(env_prefix):]
        self.supervisee_environment[env_key] = value
        LOG.info("explicitly passing env var to supervisee: %s: %s" %(env_key, value))

  def add_vars_in_config_ini_to_supervisee_environment(self, ini, ini_group):
    """ Add variables in config.ini's ini_group to supervisee's environment.
        Override already existing variables.
    """
    if ini_group not in ini.sections():
      return
    for ini_item in ini.items(ini_group):
      env_key = ini_item[0].upper()
      env_value = ini_item[1]
      action_taken = 'adding'
      if env_key in self.supervisee_environment:
        action_taken = 'overriding'
      LOG.info("config.ini %s env var to supervisee: %s: %s" %(action_taken, env_key, env_value))
      self.supervisee_environment[env_key] = env_value

  def add_cm_status_codes_to_supervisee_environment(self):
    Agent._add_cm_status_codes_to_env_dict(self.supervisee_environment)

  @staticmethod
  def _add_cm_status_codes_to_env_dict(env):
    assert(not env.has_key(CM_STATUS_CODES_ENV_VAR))
    env[CM_STATUS_CODES_ENV_VAR] = str.join(" ", CM_STATUS_CODE_SCHEMA.symbols)

  def parse_config_file(self):
    if 'CMF_CONF_DIR' in os.environ:
      conf_dir = os.environ['CMF_CONF_DIR']
    else:
      conf_dir = '/etc/cloudera-scm-agent'
    configFile = os.path.join(conf_dir, 'config.ini')
    ini = ConfigParser.SafeConfigParser()
    try:
      ini.read([configFile])
    except:
      return

    general = 'General'
    security = 'Security'
    hadoop = 'Hadoop'
    cloudera = 'Cloudera'

    try:
      self.master_host = ini.get(general, 'server_host')
    except:
      pass

    try:
      self.master_port = ini.getint(general, 'server_port')
    except:
      pass

    try:
      self.agent_httpd_port = ini.getint(general, 'listening_port')
    except:
      pass

    try:
      self.ip_address = ini.get(general, 'listening_ip')
    except:
      pass

    try:
      self.hostname = ini.get(general, 'listening_hostname')
      self.host_id = ini.get(general, 'listening_hostname')
    except:
      pass

    try:
      self.parcel_dir = ini.get(general, 'parcel_dir')
    except:
      pass

    try:
      self.use_tls = ini.getboolean(security, 'use_tls')
    except:
      pass

    try:
      self.client_key_file = ini.get(security, 'client_key_file')
    except:
      pass

    try:
      self.client_cert_file = ini.get(security, 'client_cert_file')
    except:
      pass

    try:
      self.verify_cert_file = ini.get(security, 'verify_cert_file')
    except:
      pass

    try:
      self.client_keypw_file = ini.get(security, 'client_keypw_file')
    except:
      pass

    try:
      self.optional_token = ini.getboolean(security, 'optional_token')
    except:
      pass

    self.max_collection_wait_seconds = 10.0
    try:
      self.max_collection_wait_seconds = \
          ini.getfloat(general, 'max_collection_wait_seconds')
    except:
      pass

    self.metrics_url_timeout_seconds = 30.0
    try:
      self.metrics_url_timeout_seconds = \
          ini.getfloat(general, 'metrics_url_timeout_seconds')
    except:
      pass

    self.task_metrics_timeout_seconds = 5.0
    try:
      self.task_metrics_timeout_seconds = \
          ini.getfloat(general, 'task_metrics_timeout_seconds')
    except:
      pass

    self.monitored_nodev_filesystem_types = []
    try:
      self.monitored_nodev_filesystem_types = \
          ini.get(general, 'monitored_nodev_filesystem_types').split(',')
    except:
      pass

    self.filesystem_map_poll_timeout_seconds = 2.0
    try:
      self.filesystem_map_poll_timeout_seconds = \
          ini.getfloat(general, 'filesystem_map_poll_timeout_seconds')
    except:
      pass

    self.filesystem_map_join_timeout_seconds = 0.1
    try:
      self.filesystem_map_join_timeout_seconds = \
          ini.getfloat(general, 'filesystem_map_join_timeout_seconds')
    except:
      pass

    self.filesystem_map_cache_tolerance = 60.0
    try:
      self.filesystem_map_cache_tolerance = \
          ini.getfloat(general, 'filesystem_map_cache_tolerance')
    except:
      pass

    try:
      self.supervisord_events = ini.getboolean(general, 'enable_supervisord_events')
    except:
      pass

    # Add the environment variables that start with 'CMF_AGENT_' to child process
    # environment. Then add variables from config.ini, overriding if already present.
    self.add_env_vars_with_prefix_to_supervisee_environment('CMF_AGENT_')
    self.add_vars_in_config_ini_to_supervisee_environment(ini, hadoop)
    self.add_vars_in_config_ini_to_supervisee_environment(ini, cloudera)
    self.add_cm_status_codes_to_supervisee_environment()

    try:
      # Beware that there is an inconsistency between
      # the command-line flag "--logfile" and the config
      # parameter "log_file".
      self.logfile = ini.get(general, 'log_file')
    except:
      pass

  def parse_arguments(self):
    try:
      fqdn = socket.getfqdn()
      ip_address = socket.gethostbyname(fqdn)
    except:
      LOG.exception("Could not determine hostname or ip address; proceeding.")
      fqdn = None
      ip_address = None

    default_log_dir = '/var/log/cloudera-scm-agent'
    parser = argparse.ArgumentParser(description="CMF Agent")
    parser.add_argument('--agent_dir', help="Agent's working directory.  By default, a temporary directory is used.")
    parser.add_argument('--agent_httpd_port', help="Agent's httpd port.  By default 9000 is used.", type=int, default=9000)
    parser.add_argument('--package_dir', help="Location of CMF package root.",
      required=True)
    parser.add_argument('--parcel_dir', help="Location of Parcel repository root directory.", default='/opt/cloudera/parcels')
    parser.add_argument('--supervisord_path', help="Path to supervisord",
      default=self._default_supervisord_path())
    parser.add_argument('--supervisord_httpd_port', help="Supervisor's httpd port.  By default 9001 is used.", type=int, default=9001)
    parser.add_argument('--standalone', help="Runs in standalone mode.  Heartbeats are printed to stdout. "
      + "Argument is a path to an avro data file containing a pre-set heartbeat resonse.")
    parser.add_argument('--master', help="host:port pair of master server.")
    parser.add_argument('--environment', help="KEY=VALUE[,KEY=VALUE] formatted pairs, passed to supervisee as environment")
    parser.add_argument('--host_id', help="Overrides host id of agent. Typically the host_id is the same as the hostname", default=None)
    parser.add_argument('--disable_supervisord_events', default=True, action='store_false', dest='supervisord_events',
                        help="Disable supervisord event monitoring.")
    # If could not determine fqdn, force it upon folks.
    if fqdn:
      parser.add_argument('--hostname', help="Overrides hostname of agent", default=fqdn)
      parser.add_argument('--ip_address', help="Overrides IP address of agent", default=ip_address)
    else:
      parser.add_argument('--hostname', help="Overrides hostname of agent", required=True)
      parser.add_argument('--ip_address', help="Overrides IP address of agent", required=True)

    parser.add_argument('--use_tls', help="Use TLS and certificate validation", action='store_true', default=False)
    parser.add_argument('--client_key_file', help="PEM file containing client private key", default=None)
    parser.add_argument('--client_cert_file', help="PEM file containing client certificate", default=None)
    parser.add_argument('--verify_cert_file', help="PEM file containing server verification certs", default=None)
    parser.add_argument('--client_keypw_file', help="Text file containing client private key password", default=None)
    parser.add_argument('--logfile', help="Path to rotated log file. If -, writes to stderr. Logs are rotated by default.",
        default=os.path.join(default_log_dir, "cloudera-scm-agent.log"))
    parser.add_argument('--logdir',
        help="Directory where to store sub-process log files. Default is same as agent log file, or %s if using stderr." % default_log_dir,
        default=None)
    parser.add_argument('--optional_token', help="Do not require tokens in status server.", default=False, action='store_true')
    parser.add_argument('--clear_agent_dir', help="Clear the agent dir if specified.", default=False, action='store_true')
    args = parser.parse_args()

    # Do some validation/updates
    # We get the real path (which eliminates any symlinks in the path) so that
    # later when we serve files from the agent directories we can easily
    # enforce that no symlinks are followed when restricting serving of files
    # to the process directory.
    if args.agent_dir is None:
      self.dir = os.path.realpath(tempfile.mkdtemp(prefix="cmf_agent_"))
    else:
      self.dir = os.path.realpath(args.agent_dir)
      if args.clear_agent_dir:
        LOG.info("Clearing agent dir %s" % (self.dir,))
        shutil.rmtree(self.dir)

    self.parcel_dir = args.parcel_dir
    self.agent_httpd_port = args.agent_httpd_port
    self.use_tls = args.use_tls
    self.client_key_file = args.client_key_file
    self.client_cert_file = args.client_cert_file
    self.verify_cert_file = args.verify_cert_file
    self.client_keypw_file = args.client_keypw_file
    self.logfile = args.logfile
    self.logdir = args.logdir
    if not self.logdir:
      if self.logfile != '-':
        self.logdir = os.path.dirname(self.logfile)
      else:
        self.logdir = default_log_dir
    self.supervisord_events = args.supervisord_events
    self.optional_token = args.optional_token

    self.master_host = None
    if args.master:
      self.master_host, self.master_port = args.master.split(":", 1)
      self.master_port = int(self.master_port)
    # Lazily initialized later
    self.requestor = None

    self.supervisee_environment = {}
    self.cmd_line_env_vars = {}
    if args.environment:
      for kv in args.environment.split(','):
        (k,v) = kv.split('=')
        self.cmd_line_env_vars[k] = v

    if not os.path.exists(args.supervisord_path):
      raise Exception("Supervisord path does not exist: %s" % (args.supervisord_path,))
    if not os.path.exists(args.package_dir):
      raise Exception("Package directory does not exist: %s" % (args.package_dir,))
    self.package_dir = args.package_dir
    self.supervisee_environment['CMF_PACKAGE_DIR'] = args.package_dir
    # The defaults here are similar to /etc/default/cloudera-scm-agent, except
    # that the postgresql code is done in python, and not in shell.
    mysql_jar = os.getenv('CLOUDERA_MYSQL_CONNECTOR_JAR', "/usr/share/java/mysql-connector-java.jar")
    oracle_jar = os.getenv('CLOUDERA_ORACLE_CONNECTOR_JAR', "/usr/share/java/oracle-connector-java.jar")
    cmf_mgmt_home = os.getenv('CMF_AGENT_MGMT_HOME', '/usr/share/cmf')
    if 'CMF_AGENT_MGMT_HOME' in self.cmd_line_env_vars:
      # use override for mgmt home
      cmf_mgmt_home = self.cmd_line_env_vars['CMF_AGENT_MGMT_HOME']
    postgresql_candidates = glob.glob(os.path.join(cmf_mgmt_home, "lib/postgresql-*jdbc*.jar"))
    if postgresql_candidates:
      postgresql_jar = postgresql_candidates[0]
    else:
      postgresql_jar = os.path.join(cmf_mgmt_home, "lib/postgresql-jdbc.jar")

    for jar in [mysql_jar, oracle_jar, postgresql_jar]:
      if not os.path.exists(jar):
        LOG.info("Missing database jar: %s (normal, if you're not using this database type)" % (jar,))
      else:
        LOG.info("Found database jar: %s" % (jar,))

    self.supervisee_environment['CLOUDERA_MYSQL_CONNECTOR_JAR'] = mysql_jar
    self.supervisee_environment['CLOUDERA_ORACLE_CONNECTOR_JAR'] = oracle_jar
    self.supervisee_environment['CLOUDERA_POSTGRESQL_JDBC_JAR'] = postgresql_jar

    self.supervisord_path = args.supervisord_path
    self.supervisord_httpd_port = args.supervisord_httpd_port
    self.hostname = args.hostname
    if args.host_id is None:
      self.host_id = self.hostname
    else:
      self.host_id = args.host_id
    self.ip_address = args.ip_address

    self.args = args
    LOG.info("Using directory: %s" % (self.dir,))
    LOG.info("Using supervisor binary path: %s" % (self.supervisord_path,))

  def supervisor_reload(self):
    """
    Supervisor's configuration re-reading requires
    both a re-read, and explicit starts.  supervisorctl
    seems to do this for us, whereas the XML-RPC mechanism
    requires manual work.
    """
    LOG.info("Triggering supervisord update.")
    self.supervisor_ctl.onecmd("update")
    remote_id = self.supervisor_client.supervisor.getIdentification()
    if self.identifier != remote_id:
      raise Exception("Failed to communicate to supervisor daemon.  Mismatched ids: %s vs %s" % (self.identifier, remote_id))

  def register_for_shutdown(self, f):
    """
    Calls provided callback when agent is being stopped."
    """
    self.shutdown_registry.append(f)

  def stop(self):
    if not self.started:
      return
    LOG.info("Stopping agent...")
    for f in self.shutdown_registry:
      try:
        f()
      except:
        LOG.exception("Shutdown callback failed.")
    self.started = False

  def ready_exit_on_signal(self, sig_num, stack_frame):
    """Marks loop to exit."""
    self.signal_caught = sig_num
    self.should_exit = True

  def kill_process(self, pid):
    """
    Kills the process with the given pid, first by sending SIGTERM and if
    that fails we resort to SIGKILL
    """
    # First let's ask it nicely
    try:
      os.kill(pid, signal.SIGTERM)
    except OSError, error:
      LOG.exception("Failed to kill process with pid %d" % (pid,))
      return

    # Check a few times to see if the process vanished
    for x in range(0, MAX_PROCESS_DEATH_CHECKS):
      time.sleep(PROCESS_DEATH_CHECK_INTERVAL_SECS)
      try:
        LOG.info("waiting for process to terminate...")
        os.kill(pid, 0)
      except OSError, error:
        if error.errno == errno.ESRCH:
          LOG.info("Successfully killed process with pid %d" % (pid,))
          return
      continue

    # The process is not cooperating, really kill it
    LOG.info("Killing process with pid %d" % (pid,))
    os.kill(pid, signal.SIGKILL)

class TimedHeartbeatResponse(object):
  """
  Wrapper around a heartbeat response and the heartbeat timing data
  """
  def __init__(self, response,
               ts_agent_send = None, ts_agent_recv = None,
               ts_master_send = None, ts_master_recv = None):
    self._ts_agent_send = ts_agent_send
    self._ts_agent_recv = ts_agent_recv
    self._ts_master_send = ts_master_send
    self._ts_master_recv = ts_master_recv
    self._response = response

  def get_ts_agent_send(self):
    return self._ts_agent_send

  def get_ts_agent_recv(self):
    return self._ts_agent_recv

  def get_ts_master_send(self):
    return self._ts_master_send

  def get_ts_master_recv(self):
    return self._ts_master_recv

  def get_response(self):
    return self._response

class RetryableProcessException(Exception):
  """
  An exception that describes transient errors. If
  one of these exceptions are raised, we can retry
  the operation.
  """
  def __init__(self, cause):
    self.cause = cause

  def __str__(self):
    return repr(self.cause)

class Process(object):
  """
  Manages a supervised or one-off process.
  """
  def __init__(self, agent, raw):
    self.agent = agent
    self.raw = raw
    self.id = raw["id"]
    self.name = "%s-%s" % (self.id, self.raw["name"])
    self.raw_name = self.raw["name"]
    self.dir = None
    self.logdir = None
    self.activated = False
    self.conf_path = None
    self.early_error = None
    self.has_cgroups = False

    # self.monitor may be None
    self.monitor = None
    # Monitor daemon may not exist yet if this is called very early
    if agent.monitor_daemon is not None:
      self.monitor = agent.monitor_daemon.make_monitor(raw['name'])
    else:
      LOG.warn("Not creating a monitor for %s: MonitorDaemon not created yet" %
               (self.name,))

    # TODO: Validate names?
    # TODO: Validate hostname?

  def register_error(self, error):
    LOG.debug("Registering error %s for process %s" % (error, self.name))
    if self.early_error is not None:
      LOG.warn("Overriding error message %s. Did you forget to clear_error first?" % (self.early_error,))
    self.early_error = error

  def clear_error(self):
    LOG.debug("Clearing errors for process %s" % (self.name,))
    self.early_error = None

  def refresh_monitor(self):
    if self.monitor is not None:
      self.monitor.refresh(self.dir)

  def check_user(self, user):
    try:
      pwd.getpwnam(user)
    except KeyError:
      raise Exception("User [%s] does not exist." % (user,))

  def check_group(self, group):
    try:
      grp.getgrnam(group)
    except KeyError:
      raise Exception("Group [%s] does not exist." % (group,))

  def check_user_and_group(self, user, group):
    self.check_user(user)
    self.check_group(group)

  def activate(self):
    if self.activated:
      # Re-evaluate any dynamic resources.
      self.evaluate_resources(False)
    else:
      LOG.info("Activating Process %s" % (str(self),))
      # Create working dir
      user = self.raw["user"]
      group = self.raw["group"]

      # This is a common problem so we do an explicit check
      self.check_user_and_group(user, group)

      self.dir = self.agent.mkworkdir("process", self.name, user=user, group=group)

      self.write_process_conf()
      self.borrow_keytab()

      # Create log directory for the process
      self.logdir = self.agent.mkworkdir("process", self.name, "logs", user=user, group=group)

      # Control group creation is conditioned on the following:
      # 1. Whether RM is enabled. Just because we have mounted cgroup
      #    subsystems doesn't mean we should create new cgroups.
      # 2. Whether the process is one off. The lifecycles of one off
      #    processes are radically different: they may never deactivate.
      if self.agent.rm_enabled and not self.raw["one_off"]:
        self.has_cgroups = self.agent.cg_manager.create_group(self.name)

      self.evaluate_resources(True)

      first_time = self.conf_path is None;

      # Write supervisor configurations (this sets self.conf_path)
      self.write_conf()
      self.agent.supervisor_reload()

      # Set the selected parcels for this process as 'active' at the host level.
      # This is a semantic mismatch as different processes could be using
      # different versions of the same product, but we forbid this at higher
      # levels. If we truly expose per-process parcels, then the active concept
      # would go completely out the window, and we'd have to discard this.
      self.agent.repo.set_active_parcels(self.raw['parcels'])

      # Only call startProcess if this is not the first time we've exposed
      # the process to supervisor. Otherwise supervisor_reload() will have
      # caused supervisor to start the process, and should the process finish
      # quickly, startProcess might accidentally start it again.
      if not first_time:
        try:
          # Reloading a process configuration that already exists will not
          # change the process state, so we have to start it explicitly.
          self.agent.supervisor_client.supervisor.startProcess(self.name)
        except Fault, fault:
          if fault.faultCode != Faults.STILL_RUNNING and fault.faultCode != Faults.ALREADY_STARTED:
            raise RetryableProcessException(fault)

      self.refresh_monitor()
      self.activated = True

  def write_process_conf(self, namelist=None):
      user = self.raw["user"]
      group = self.raw["group"]

      if self.agent.honor_users_and_groups:
        # This is a common problem so we do an explicit check
        self.check_user_and_group(user, group)
        uid, gid = user_group_to_ids(user, group)
      else:
        uid, gid = None, None

      # Even if uid/gid are not None, they are only honored when unzipping
      # new files.
      unzip(self.raw["configuration_data"], self.dir, uid=uid, gid=gid,
            namelist=namelist)

      # Broaden permissions on everything but keytab files. Files and
      # directories created in unzip have mode 600 (files) and 700
      # (directories). Extend them to 640 (files) and 751 (directories).
      #
      # Note that if an explicit list of names was specified, the changes
      # are restricted to only those files and directories in the list.
      #
      # TODO: Ideally each file's mode would be defined in configuration_data,
      # but the ZIP archive format doesn't support permissions.
      for root, dirs, files in os.walk(self.dir):
        for f in files:
          if not f.endswith(".keytab") and (namelist == None or f in namelist):
            os.chmod(os.path.join(root, f), 0640)

        for d in dirs:
          if namelist == None or d in namelist:
            os.chmod(os.path.join(root, d), 0751)

      # For a supervised process, all the config files in the config dir.
      # However, some files need to be deployed to special locations and
      # need to be owned by a user other than the one running the process.
      # Deploy them now, provided they're in the explicit list of names (if
      # one exists).
      if self.raw["special_file_info"]:
        # We include both the inherited environment (os.environ)
        # and the additional environment.
        env = os.environ.copy()
        env.update(self.get_process_environment())
        # To correctly support parcel deployments, we must obtain the parcels'
        # environment variables too. This has to be done through a shell as the
        # parcels provide their environment in the form of a shell script.
        env.update(source(self.resolve_program("common/cloudera-config.sh"),
                          "source_parcel_environment", env))
        for info in self.raw["special_file_info"]:
          try:
            if namelist == None or info["name"] in namelist:
              deploy_special_file(info, self.dir, env,
                                  self.agent.honor_users_and_groups)
          except Exception, e:
            LOG.exception("Error while deploying: %s.  Environment was: %s." % (info, env))


  def borrow_keytab(self):
    """We borrow the process credential for monitoring. OPSAPS-6906."""
    # Always use the HTTP principal.
    #
    # The process's krb principal is hardcoded by CM (always `hdfs', `mapred',
    # etc.). And that krb principal is not the same as the process user, which
    # is configurable. So the conclusion is that we have to do some hardcoding.
    # And that might as well be `HTTP'.
    princ_prefix = "HTTP"

    for root, dirs, files in os.walk(self.dir):
      for f in files:
        if f.endswith(".keytab"):
          f_path = os.path.join(root, f)

          # Is our principal potentially in there?
          if princ_prefix not in file(f_path).read():
            continue

          cmf.kt_renewer.async_kinit_from_keytab(
              f_path, "%s/%s" % (princ_prefix, self.agent.hostname,))


  def get_process_environment(self):
    environment = self.agent.supervisee_environment.copy()
    environment['CONF_DIR'] = self.dir

    self.agent.repo.prepare_environment(self.raw['parcels'], environment)

    raw_env = self.raw["environment"]
    if 'SET_PYTHON_PATH' in raw_env:
      environment['PYTHONPATH'] = ":".join(sys.path)
    for key, value in raw_env.iteritems():
      environment[key] = value
    return environment

  def write_conf(self):
    environment = self.get_process_environment()

    sup_data = {
      "program_name": self.name,
      "command": self.resolve_program(self.raw["program"]),
      "arguments": self.raw["arguments"],
      "directory": self.dir,
      "environment": environment,
      "one_off": self.raw["one_off"],
      "auto_restart": self.raw["auto_restart"],
      "exitcodes": ",".join(str(EC) for EC in SUPERVISOR_EXPECTED_EXITCODES)
    }
    if self.agent.honor_users_and_groups:
      sup_data["user"] = self.raw["user"]
      sup_data["group"] = self.raw["group"]

    if self.has_cgroups:
      sup_data["cgroups"] = ",".join(self.agent.cg_manager.get_groups(self.name))

    sup_config = render_text("supervisee.conf", data=sup_data)

    f = file(os.path.join(self.agent.supervisor_include_path, self.name + ".conf"), "w")
    f.write(sup_config)
    f.close()
    self.conf_path = f.name

  def resolve_program(self, program):
    """
    Returns path of program relative to package_dir,
    if program is a relative path.
    """
    if program.startswith("/"):
      return program
    else:
      return os.path.join(self.agent.package_dir, program)

  def deactivate(self, forget):
    if self.monitor is not None:
      self.monitor.stop()

    needs_reload = False

    if self.activated == True:
      LOG.info("Deactivating process %s" % (str(self),))
      try:
        self.agent.supervisor_client.supervisor.stopProcess(self.name)
        needs_reload = True
      except Fault, fault:
        # NOT_RUNNING: It's already stopped, so why worry
        # BAD_NAME: Supervisord doesn't know what we're on about.
        #           Even if it's running, we can't help.
        if fault.faultCode == Faults.NOT_RUNNING:
          pass
        elif fault.faultCode == Faults.BAD_NAME:
          LOG.info("Supervisord doesn't know about process %s. "
                   "It cannot be deactivated by this Agent, "
                   "assuming it even exists" % (str(self),))
        else:
          raise RetryableProcessException(fault)

    if forget == True:
      LOG.info("Deleting process %s" % (str(self),))
      if self.conf_path is not None:
        needs_reload = True
        self.delete_conf()
        self.conf_path = None

    # Ignore the value of 'activated': some processes (such as HBase)
    # won't be deactivated, they'll just be forgotten the next time
    # they start up. Because we unset 'has_cgroups', this is safe.
    if self.has_cgroups:
      self.agent.cg_manager.destroy_group(self.name)
      self.has_cgroups = False

    # We only need to talk to supervisord when
    # we've encountered a substantive change.
    if needs_reload:
      self.agent.supervisor_reload()

    self.activated = False

  def list_confdir(self):
    """
    Returns a sorted list of files contained in the process's
    private directory.
    """
    if self.dir is None:
      return []
    listing = os.listdir(self.dir)
    listing.sort()
    processdir = os.path.basename(self.dir)
    return ["/".join(["process", processdir, "files", file]) for file in listing if file != 'logs']

  def list_logdir(self):
    """
    Returns a sorted list of log files contained in the process's
    private log directory.
    """
    if self.logdir is None:
      return []
    listing = os.listdir(self.logdir)
    listing.sort()
    processdir = os.path.basename(self.dir)
    return ["/".join(["process", processdir, "files", "logs", file]) for file in listing]

  def delete_conf(self):
    os.remove(self.conf_path)

  def evaluate_resources(self, first_time):
    """
    Evaluate process resources, making system changes as needed. Unknown
    or unused resources are ignored.

    If first_time is set, will evaluate both static and dynamic resources.
    """

    for res in self.raw["resources"]:
      # Skip static resources if it's not the first time.
      if not res["dynamic"] and not first_time:
        continue

      try:
        # Create directories.
        d = res["directory"]
        if d:
          self.agent.mkabsdir(path=d["path"], user=d["user"],
                              group=d["group"], mode=d["mode"])

        # Set cgroup cpu shares.
        cs = res["cpu"]
        if cs and self.has_cgroups:
          self.agent.cg_manager.configure_group(self.name, "cpu", cs["shares"])

        # Set cgroup io weight.
        io = res["io"]
        if io and self.has_cgroups:
          self.agent.cg_manager.configure_group(self.name, "blkio", io["weight"])

        # Set cgroup memory hard limit.
        mem = res["memory"]
        if mem and self.has_cgroups:
          val = mem["hard_limit"]
          if val == -1:
            # Special value -1 means max long (2^63 - 1).
            val = 0x7fffffffffffffffL
          else:
            # Convert from megabytes to bytes.
            val *= (1024 * 1024)
          self.agent.cg_manager.configure_group(self.name, "memory", val)
      except:
        # Resource evaluation failures are non-fatal.
        LOG.exception("Could not evaluate resource %s" % (res,))

  def make_status(self, supervisor_data):
    """
    Returns a ProcessStatus object given supervisor_data.
    """
    process_status = dict()
    process_status["run_generation"] = self.raw["run_generation"]
    process_status["config_generation"] = self.raw["config_generation"]

    process_stats = dict()
    process_stats["id"] = self.raw["id"]
    # See OPSAPS-4593. We don't report most stats at the moment, and these are
    # not displayed anywhere in our UI. The schema and code here remain since
    # we plan to add these back shortly.
    process_stats["rss"] = -1
    process_stats["vms"] = -1
    process_stats["cpu_percent"] = -1.0
    process_stats["user_cpu_time"] = -1.0
    process_stats["system_cpu_time"] = -1.0

    pid = None

    process_status["name"] = self.raw["name"]

    data = None
    if supervisor_data is not None:
      data = supervisor_data.get(self.name)
    if self.early_error is not None:
      fail_str = "Error found before invoking supervisord: %s" % self.early_error
      process_status["status"] = "FATAL"
      process_status["stdout_tail"] = fail_str
      process_status["stderr_tail"] = fail_str
      process_status["cm_status_code"] = ""
      process_status["exit_code"] = -1
      process_status["pid"] = -1
      process_stats["wallclock_time"] = -1
    elif data is None:
      # If supervisor doesn't know about the process, it's certainly
      # not running in our view of the world. This is a more useful
      # response in edge cases where the Agent is restarted and forgets
      # about proccesses.
      #
      # However, if we had problems talking to supervisor, respond with
      # an unknown status. This has been observed when the machine is
      # under heavy load and can be safely considered a transient state
      # because if it happens often enough, the agent will have exited.
      process_status["pid"] = -1
      if supervisor_data is None:
        process_status["status"] = "UNKNOWN"
      else:
        process_status["status"] = "STOPPED"
      process_status["stdout_tail"] = "UNKNOWN"
      process_status["stderr_tail"] = "UNKNOWN"
      process_status["cm_status_code"] = ""
      process_status["exit_code"] = -1
      process_stats["wallclock_time"] = -1
    else:
      pid = data["pid"]
      process_status["pid"] = pid

      # Heuristic to protect against OPSAPS-9208.
      #
      # If an auto-restartable process fails with an unexpected exit code,
      # supervisor will transition it to EXITED, and then start it up again
      # (transitioning it to STARTING). We must be careful not to publish
      # the momentary EXITED state; otherwise it might cause an in-flight
      # bring up command to fail.
      #
      # In effect, this heuristic makes automatic restarts more invisible
      # to the system (though not completely; the subsequent STARTING state
      # will still be published).
      #
      # Note that an auto restarted process cannot be one off; this is
      # guaranteed by supervisee.conf.
      state_name = data["statename"]
      exit_code = data.get("exitstatus", -1)
      if state_name == "EXITED" and \
            self.raw["auto_restart"] and \
            exit_code not in SUPERVISOR_EXPECTED_EXITCODES:
        state_name = "STARTING"

      process_status["status"] = state_name
      process_status["stdout_tail"] = try_tail(data["stdout_logfile"], DEFAULT_LOG_FILE_TAIL)
      process_status["stderr_tail"], process_status["cm_status_code"] = \
          Process._extract_err_str_and_cm_status_code(try_tail(data["stderr_logfile"], DEFAULT_LOG_FILE_TAIL))
      process_status["exit_code"] = exit_code
      if data["stop"] == 0:
        process_stats["wallclock_time"] = data["now"] - data["start"]
      else:
        process_stats["wallclock_time"] = data["stop"] - data["start"]

    return process_status, process_stats

  @staticmethod
  def _extract_err_str_and_cm_status_code(err_log):
    if err_log is None:
      return "", ""

    search_start = max(0, len(err_log) - MAX_CM_STATUS_CODE_STR_LEN)
    match = re.search(CM_STATUS_CODE_REGEX, err_log[search_start:])
    if match is not None:
      # "- 1" required below to skip the new line character preceding the error
      # code.
      return err_log[:max(0, search_start + match.start() - 1)], match.group(1)
    else:
      return err_log, ""

  def __str__(self):
    return self.name

LOG_FORMAT = '[%(asctime)s] %(process)d %(threadName)s %(module)-12s %(levelname)-8s %(message)s'
DATE_FORMAT = '%d/%b/%Y %H:%M:%S %z'
BUFFER_SIZE = 10*1024*1024 # This is the size in characters, not bytes

def close_most_fds():
  """
  Closes most file descriptors (i.e. all but stdin, stdout, and stderr).
  They may have been inherited from the parent process and, if they're pipes,
  will prevent the parent process from exiting.

  See the following URL for more details:
  http://code.activestate.com/recipes/278731-creating-a-daemon-the-python-way
  """
  if os.uname()[0] == "Darwin":
    # OS X Snow Leopard throws Illegal Instruction here.
    # It's the same bug probably as referred to by
    # http://lists.supervisord.org/pipermail/supervisor-users/2010-January/000525.html
    # Note: we don't really support Darwin as a target platform,
    # but sometimes it's handy for development.
    return

  maxfd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
  if (maxfd == resource.RLIM_INFINITY):
    maxfd = 1024

  # Close the fds. We'll get an OSError if a particular fd was already closed.
  for fd in range(3, maxfd):
    try:
      os.close(fd)
    except OSError:
      pass

def setup_initial_logging():
  # We don't know the initial log configuration
  # right away, so we use a basic log configuration
  # to start with.
  if os.environ.get('SCM_DEBUG') is not None:
    level = logging.DEBUG
  else:
    level = logging.INFO

  logging.basicConfig(level=level,
    format=LOG_FORMAT, datefmt=DATE_FORMAT)
  LOG.info("SCM Agent Version: %s" % (cmf.version.VERSION))

def setup_rolling_logging(filename):
  # Create the directory, if it doesn't already exist.
  filename_dir = os.path.dirname(filename)
  if not os.path.exists(filename_dir):
    try:
      os.makedirs(filename_dir)
      LOG.info("Created log directory: %s" % (filename_dir, ))
    except:
      LOG.exception("Failed to create log directory: %s" % (filename_dir, ))

  # Now that we have more information, we setup
  # rolling logging.
  root_logger = logging.getLogger()
  assert 1 == len(root_logger.handlers)
  original_handler = root_logger.handlers[0]

  # Buffer to keep recent log messages
  buffer_handler = FixedBufferHandler(BUFFER_SIZE)
  formatter = logging.Formatter(LOG_FORMAT, DATE_FORMAT)
  # We always want to catch all messages in our error report buffer
  buffer_handler.setLevel(logging.DEBUG)
  buffer_handler.setFormatter(formatter)
  root_logger = logging.getLogger()
  root_logger.addHandler(buffer_handler)

  # Rotating file handler, unless filename is "-"
  if filename != "-":
    try:
      rh = logging.handlers.RotatingFileHandler(filename,
        maxBytes=MAX_LOG_SIZE_BYTES, backupCount=LOG_BACKUP_COUNT)
      rh.setFormatter(formatter)
      logging.info("Logging to " + filename)
      root_logger.addHandler(rh)
      root_logger.removeHandler(original_handler)
    except:
      logging.exception("Could not open log file; writing to stderr.")

def get_log_buffer():
  l = logging.getLogger()
  for h in l.handlers:
    if isinstance(h, cmf.log_buffer.FixedBufferHandler):
      return h.buf
  else:
    return None

def get_log_rotating_file_handler():
  l = logging.getLogger()
  for h in l.handlers:
    if isinstance(h, logging.handlers.RotatingFileHandler):
      return h
  return None

def try_tail(filename, length):
  """
  If, for whatever reason, the log file is deleted
  or unavailable, we want to degrade gracefully.
  """
  try:
    return tail(filename, length)
  except:
    LOG.exception("Could not tail file: %s" % (filename,))
    return "Failed to read log."

def generate_random_bytes(num):
  return "".join(chr(random.randint(0,255)) for i in xrange(num))

def main():
  close_most_fds()
  setup_initial_logging()
  register_sigquit_stack_dump()
  agent = Agent()
  agent.start()

if __name__ == "__main__":
  # An environment variable can enable profiling.
  if os.environ.get("CLOUDERA_SCM_AGENT_PROFILE") is not None:
    try:
      # python >= 2.5
      import cProfile as profile
    except:
      import profile as profile
    profiling_filename = "/tmp/agent_profile." + str(time.time())
    # Logging is not set up yet, so we just print to stdout.
    print "Profiling to: " + profiling_filename
    print "Read (after killing process) with 'import pstats; pstats.Stats('%s').sort_stats('time').print_stats(60)'" % (profiling_filename,)
    profile.run('main()', '/tmp/agent_profile.' + str(time.time()))
  else:
    main()
