# Copyright (c) 2011 Cloudera, Inc. All rights reserved.

import logging
import psutil
import simplejson
import threading
import time
import traceback

from cmf.monitor import schema, constants
from cmf.monitor.base import MonitorInterface
from cmf.monitor.dir_monitor import DirMonitor
from cmf.monitor.monitor_properties import ConfParser
from cmf.monitor.wakeable_thread import WakeableThread

from cmf.url_util import urlopen_with_timeout

LOG = logging.getLogger(__name__)

_COLLECT_INTERVAL = 'collect_interval'

class CollectionError(Exception):
  def __init__(self, code):
    self.code = code;

class AbstractMonitor(MonitorInterface):
  """
  A common implementation of the MonitorInterface. Expectation on a derived
  class:
    * It uses a separate thread to do metrics collection, and implements
      the `collect()' method to perform the collection work.
    * It checks for configuration update (captured in `conf_parser') before
      each collection.
    * The first config section should contain a `collect_interval'.
  """

  def __init__(self, daemon, registry):
    MonitorInterface.__init__(self)
    self._daemon = daemon
    self._registry = registry
    self._collection_thread = None
    self._conf_parser = None
    self._last_error = None
    self._last_reported_timestamp = None

    # If the _pid is None then it means that the associated process is not
    # running, and that metric collection dependent on that process should
    # be skipped.
    self._pid = None

    # We throttle our missing pid log messages. See OPSAPS-5138.
    self._last_missing_pid = None

    # If _unexpected_exits is 0 then it means that the associated
    # process has not encountered any unexpected exits since the
    # last role update.
    self._unexpected_exits = 0

    # We use a lock to ensure that read and write accesses to
    # _unexpected_exits for instance don't step on each other.
    self.lock = threading.RLock()

  def collect(self):
    """
    The loop content in the collection thread.
    This method is responsible for loading any new update in `conf_parser'.
    """
    # Should be overridden by derived classes
    raise NotImplementedError("%s needs to override collect()" %
                              (self.__class__,))

  @property
  def conf_parser(self):
    return self._conf_parser

  @property
  def last_error(self):
    return self._last_error

  @property
  def daemon(self):
    return self._daemon

  def record_error(self, exc):
    self._last_error = "%s\n%s" % (exc, traceback.format_exc())

  def refresh(self, conf_dir):
    """
    Refresh the monitor context and start the monitor.
    The monitor configuration is loaded into `conf_parser' after each refresh.
    """
    LOG.info("Refreshing %s" % (self,))
    try:
      self._conf_parser = ConfParser(conf_dir)
    except IOError, ex:
      self.record_error(ex)
      LOG.error("%s: %s" % (self, ex))
      return

    self.ensure_collecting()

  def refresh_pid(self, pid, skip_pid_checks=False):
    # If we have not been given a pid to monitor, or we've been instructed to
    # skip our pid checks (for testing) we're done.
    if pid is None or skip_pid_checks:
      self._set_pid(pid)
      return

    # If we have been given the same pid we're already monitoring, then we have
    # nothing to do.
    if pid == self._pid:
      return

    # First we check if the pid we've been given is for a process that looks
    # like what we're supposed to be monitoring. It's possible that it is not,
    # for example it might be an su process that is the parent of the actual
    # process we're interested in monitoring. See OPSAPS-4451.
    process = psutil.Process(pid)
    if self.is_process_to_monitor(process):
      self._set_pid(pid)
      return

    # If the pid we were given was not the process we're looking for, we'll
    # look through each of it's children looking for an appropriate process.
    for child in process.get_children():
      if self.is_process_to_monitor(child):
        self._set_pid(pid)
        return

    # If we found no match, then we turn off the monitor.
    self._pid = None
    if pid != self._last_missing_pid:
      LOG.error("%s unable to find appropriate process to monitor for pid %d." % 
                (self, pid))
      self._last_missing_pid = pid

  def record_unexpected_exit(self):
    try:
      self.lock.acquire()
      self._unexpected_exits += 1
    finally:
      self.lock.release()

  def add_unexpected_exit_metrics(self, update):
    """
    Adds the current unexpected exit count as a metric to the (role) update
    and then resets the count.
    """
    try:
      self.lock.acquire()
      update.add_metric(schema.UNEXPECTED_EXITS, self._unexpected_exits)
      self._unexpected_exits = 0
    finally:
      self.lock.release()

  def _set_pid(self, pid):
    self._pid = pid
    self._last_missing_pid = None

  def is_process_to_monitor(self, process):
    if process is None:
      return False
    # We never want to monitor the priveleged 'jsvc.exec' process. We usually
    # want to monitor its child 'jsvc.exec' process which will not be running
    # as root.
    if process.name == 'jsvc.exec' and process.uids[0] == 0:
      return False
    # As of this moment, the only processes we monitor are either 'java' or
    # 'jsvc.exec'. This check is sufficient to correctly identify our current
    # monitors, but may need to evolve in the future as we add more monitors.
    # With this check, we will not monitor 'su' processes, nor will we monitor
    # the wrong part of the 'jsvc.exec' process. Also, we will not end up
    # monitoring various 'tail' processes started by 'jsvc.exec'. We originally
    # tried to do more here -- matching particular arguments on the command
    # line -- but ran into issues when that command line regularly exceeded
    # the 4K characters exposed by /proc.
    return process.name in ('java', 'jsvc.exec')

  def ensure_collecting(self):
    if self._collection_thread is not None:
      return

    self._collection_thread = self.make_collection_thread()
    self._collection_thread.start()
    self._registry.register(self)
    self._daemon.new_monitor(self)

  def make_collection_thread(self):
    return WakeableThread(
        "Monitor-%s" % (self.__class__.__name__,),
        self.collect)

  def _get_collect_interval(self):
    val = self._conf_parser.getfloat(self._conf_parser.get_default_section(),
                                     _COLLECT_INTERVAL)
    if val is None:
      LOG.warn("%s is missing the '%s' config" %
               (self._conf_parser.path, _COLLECT_INTERVAL))
      return self._daemon.get_interval()
    return val

  def stop(self):
    """Stop the monitor"""
    self._registry.unregister(self)
    if self._collection_thread is not None:
      self._collection_thread.should_stop()
      self._collection_thread = None

  def get_attempt_updates(self):
    """Returns a list of avro'd AttemptUpdate objects to be sent to the firehose"""
    return []

  def get_role_update(self):
    """Returns a RoleUpdate to be sent to the firehose"""
    return None

  def get_service_update(self):
    """Returns a ServiceUpdate to be sent to the firehose"""
    return None

  def start_collecting(self, latch):
    if self._collection_thread is not None:
      self._collection_thread.wake(latch)

  def urlopen(self, url):
    # We always pass the secure http service name because it has no effect on
    # unsecured connections.
    return urlopen_with_timeout(url,
                                timeout=self.get_metrics_url_timeout_seconds(),
                                secure_http_service_name=self._get_secure_http_service_name())
 
  def _get_secure_http_service_name(self):
    """
    Constructs a service name for a secure HTTP service on this host. This is based 
    on the format of the urllib2_kerberos library which, by default, builds the
    service name with the host from the HTTP request. The service name is used
    by GSS to locate the secure service. While we know the service to be Kerberos-
    based, the GSS API is more generic and doesn't call these Kerberos principals,
    so we refer to them by the GSS term 'service name'.
    """
    return "HTTP@%s" % (self._daemon.get_agent().hostname,)

  def collect_metrics_from_url(self, url):
    result = None
    start = time.time()
    try:
      code = constants.WEB_METRICS_COLLECTION_COMMUNICATION_FAILURE
      openedUrl = self.urlopen(url)
      code = constants.WEB_METRICS_COLLECTION_PARSE_FAILURE
      result = self.parse_metrics(simplejson.load(openedUrl))
    except Exception, ex:
      self.record_error(ex)
      LOG.exception("Error fetching metrics at '%s'" % (url,))
      result = CollectionError(code)
    now = time.time()
    duration = int((now - start) * 1000.0)
    return (result, now, duration)

  def get_task_metrics_timeout_seconds(self):
    return self._daemon.get_task_metrics_timeout_seconds()

  def get_metrics_url_timeout_seconds(self):
    return self._daemon.get_metrics_url_timeout_seconds()

  # For testing. 
  def clear_last_reported_timestamp(self):
    self._last_reported_timestamp = None

  # This method should be overridden by monitors to control when they are ready
  # to report metrics. This basic implementation says that the monitor is ready
  # as soon as the process pid is set. Examples of when this is not appropriate
  # are the host monitor, which has no PID, and something like a DataNode which
  # does have a pid, but also has a web UI from which it collects metrics that
  # it wants to ensure is up before claiming to be ready.
  #
  # Note that this is called by the overall monitoring daemon scheduling thread
  # so this needs to return quickly (calls should take no more than a second)
  # and it needs to be mindful of what members are accessed.
  def is_ready_to_report(self):
    return self._pid is not None

  # Returns true if the input URL responded with some JSON string within the
  # input timeout. False otherwise
  def is_metrics_url_responding(self, url, timeout):
    try:
      simplejson.load(urlopen_with_timeout(url, timeout=timeout))
      return True
    except Exception, ex:
      pass
    return False

  def create_dir_monitor(self, directory, monitors):
    try:
      monitors.append(DirMonitor(self._daemon.get_filesystem_map(),
                                 directory))
    except Exception:
      # This might happen if the directory for which we are trying to
      # instantiate a monitor for is mounted on a network share. The
      # DirMonitor uses psutil that only gives information for local file
      # systems. We should probably generate a health warning for this,
      # OPSAPS-4139 tracks this.
      LOG.exception("Could not create directory monitor for %s" % 
                    (directory,))

  def add_metrics_from_webserver(self,
                                 metrics_tuple,
                                 last_reported_timestamp,
                                 update):
    metrics, timestamp, duration = metrics_tuple
    if metrics is None or timestamp == last_reported_timestamp:
      return False

    update.add_metric(schema.WEB_METRICS_COLLECTION_DURATION, duration)
    update.timestamp = timestamp
    if isinstance(metrics, CollectionError):
      update.add_metric(schema.WEB_METRICS_COLLECTION_STATUS, metrics.code)
    else:
      update.add_metric(schema.WEB_METRICS_COLLECTION_STATUS,
                        constants.WEB_METRICS_COLLECTION_GOOD)
      self.add_webserver_metrics_to_role_update(update, metrics)
    return True

  # This method should be overridden by monitors to add metrics from the input
  # JSON metrics structure, retrieved from the monitored role's webserver, to
  # the input role update
  def add_webserver_metrics_to_role_update(self, update, metrics):
    pass
